/*
 * icx_lcd_tovis_ILI9327.c
 *
 * Copyright 2011-2015 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>
#include <asm/io.h>

#ifdef CONFIG_WM

#include <asm/errno.h>

#include "reg.h"
#include "icx_io.h"
#include "icx_smu.h"
#include "icx_gpio.h"
#include "icx_spi.h"
#include "icx_pwc.h"
#include "icx_pwm.h"
#include "icx_lcdc.h"
#include "icx_lcd.h"


/*******************/
/*@ LCDC operation */
/*******************/

static int  lcd_tovis_ili9327_power_on(const lcdc_config_t* config);
static void lcd_tovis_ili9327_power_off(void);

const lcd_ops_t lcd_tovis_ili9327_ops = {
	.power_on  = lcd_tovis_ili9327_power_on,
	.power_off = lcd_tovis_ili9327_power_off,
};

/*****************/
/*@ LCDC setting */
/*****************/

/* hsync */
#define LCD_ILI9327_HAREA		LCDC_WIDTH_WQVGA
#define	LCD_ILI9327_HPULSE		2
#define	LCD_ILI9327_HFRONTP		6
#define	LCD_ILI9327_HBACKP		8
#define	LCD_ILI9327_HBLANK		(LCD_ILI9327_HFRONTP + LCD_ILI9327_HPULSE + LCD_ILI9327_HBACKP)
#define	LCD_ILI9327_HTOTAL		(LCD_ILI9327_HAREA   + LCD_ILI9327_HBLANK)
#define	LCD_ILI9327_HEDGE1		(LCD_ILI9327_HFRONTP)
#define	LCD_ILI9327_HEDGE2		(LCD_ILI9327_HFRONTP + LCD_ILI9327_HPULSE)

/* vsync */
#define LCD_ILI9327_VAREA		LCDC_HEIGHT_WQVGA
#define	LCD_ILI9327_VPULSE		2
#define	LCD_ILI9327_VFRONTP		2
#define	LCD_ILI9327_VBACKP		2
#define	LCD_ILI9327_VBLANK		(LCD_ILI9327_VFRONTP + LCD_ILI9327_VPULSE + LCD_ILI9327_VBACKP)
#define	LCD_ILI9327_VTOTAL		(LCD_ILI9327_VAREA   + LCD_ILI9327_VBLANK)
#define	LCD_ILI9327_VEDGE1		(LCD_ILI9327_VFRONTP)
#define	LCD_ILI9327_VEDGE2		(LCD_ILI9327_VFRONTP + LCD_ILI9327_VPULSE)

/* default brightness */
#define	LCD_ILI9327_PWM_DEFAULT	(PWM_TOTAL_100K / 2)

/* LCDC config */
const lcdc_config_t lcdc_tovis_ili9327_config = {
	.width	  = LCDC_WIDTH_WQVGA,
	.height   = LCDC_HEIGHT_WQVGA,

	.clkdiv   = SMU_PLLSEL_PLL3 | SMU_DIV(36),

	.frameadr = LCDC_FRAMEADR_DEFAULT,
	.areaadr  = LCDC_AREAADR_DEFAULT,
	.pol      = (LCDC_CLKPOL_POS | LCDC_HPOL_POS | LCDC_VPOL_POS | LCDC_ENPOL_HI),
	.htotal   = LCD_ILI9327_HTOTAL,
	.harea	  = LCD_ILI9327_HAREA,
	.hedge1   = LCD_ILI9327_HEDGE1,
	.hedge2   = LCD_ILI9327_HEDGE2,
	.vtotal   = LCD_ILI9327_VTOTAL,
	.varea    = LCD_ILI9327_VAREA,
	.vedge1   = LCD_ILI9327_VEDGE1,
	.vedge2   = LCD_ILI9327_VEDGE2,

	.backlight = {
		.dev  = PWM_DEV_1,
		.pol  = LCDC_BLPOL_NEG,
		.full = LCD_ILI9327_PWM_DEFAULT,
		.div  = SMU_PLLSEL_PLL3 | SMU_DIV(4),
	},
};

/****************/
/*@ SPI setting */
/****************/

/* SPI config */
static spi_config_t spi_config_w = {
	.dev = SPI_DEV_SP3,
	.clk = SPI_CLK_6MHZ,
	.cs  = SPI_CS_SEL_CS0,
	.pol = (SPI_CK_DLY_ON | SPI_CK_POL_POS | SPI_CS_POL_NEG),
	.nbw = SPI_NB_9BIT,
	.nbr = 0,
};

/* SPI start byte */
#define LCD_ILI9327_SB_CMD	0x0000          /* Set index register                           */
#define LCD_ILI9327_SB_DATA    0x0100          /* Writes instruction or RAM data       */

#define LCD_ILI9327_IR_SYNC	0xFFFE		/* transfer synchronize(local rule)	*/
#define LCD_ILI9327_IR_WAIT	0xFFFF		/* wait command(local rule)			*/

#define	LCD_ILI9327_HEART_BEAT_MAGIC_NUMBER	0xAA55

/* command type */
typedef struct lcd_ili9327_command {
	const unsigned short    ir;
	union {
		const unsigned short    data_size;
		const unsigned short    wait_time_ms;
	};
	const char* const       pdata;
} lcd_ili9327_command_t;

/* SPI command */
static const lcd_ili9327_command_t spi_cmd_power_on[] = {
	/* Power Setting */
	{ 0xD1, { 3 }, "\x00\x6B\x13" },
	{ 0xD0, { 3 }, "\x07\x04\x91" },

	/* Hidden Code */
	{ 0xE9, { 1 }, "\x20" },
	{ 0xE5, { 2 }, "\x91\xB1" },

	/* Display Control setting */
	{ 0xB4, { 1 }, "\x11" },
	{ 0xC6, { 1 }, "\x02" },
	{ 0x36, { 1 }, "\x40" },
	{ 0x3A, { 1 }, "\x60" },
	{ 0xC1, { 4 }, "\x10\x10\x02\x02" },
	{ 0xC0, { 6 }, "\x12\x35\x00\x00\x01\x02" },
	{ 0xD2, { 2 }, "\x01\x44" },
	{ 0xB3, { 4 }, "\x02\x00\x00\x30" },

	/* Gamma Settings */
	{ 0xC8, { 15 }, "\x00\x77\x77\x13\x04\x07\x00\x00\x67\x10\x02\x00\x00\x00\x60" }, /* p1 - p15 */

	/* Sleep Out */
	{ 0x11, { 0 }, NULL },
	{ LCD_ILI9327_IR_WAIT, { 120 }, NULL },

	/* Display On */
	{ 0x29, { 0 }, NULL },
};
#define SPI_CMD_POWER_ON_NUM    ARRAY_SIZE(spi_cmd_power_on)

#define SPI_CMD_DATA_MAX_SIZE   (1 + 15) /* command(1) + data(15) */

static const lcd_ili9327_command_t spi_cmd_power_off[] = {

	/* Display Off */
	{ 0x28, { 0 }, NULL },

	 /* sleep in */
	{ 0x10, { 0 }, NULL },
	{ LCD_ILI9327_IR_WAIT, { 60 }, NULL },
};

#define SPI_CMD_POWER_OFF_NUM   ARRAY_SIZE(spi_cmd_power_off)

/******************/
/*@ Local routine */
/******************/

/*-----------------------------------------------------------------------------
 * Function : lcd_ili9327_send_command
 * Abstruct : Send a command & a data to LCD via SPI
 * Input    : none
 * Return   : <= 0(NG) / others(sent packet=success)
 *----------------------------------------------------------------------------*/
static int lcd_ili9327_send_command(const lcd_ili9327_command_t* func)
{
	int count = 0;
	int ret   = 0;
	u_int buf[SPI_CMD_DATA_MAX_SIZE] = { 0x0 };

	if (func->ir == (unsigned short)LCD_ILI9327_IR_WAIT) {
		/* wait */
		mdelay(func->wait_time_ms);
		ret = func->wait_time_ms;
	}
	else {
		/* make command and data */
		buf[count] = LCD_ILI9327_SB_CMD | func->ir;
		++count;
		if (func->data_size && func->pdata) {

			int i;

			if (SPI_CMD_DATA_MAX_SIZE < func->data_size) {
				printf("%s(): error -1\n", __FUNCTION__);

				return -1;
			}

			for (i = 0; i < func->data_size; ++i) {
				buf[count] = LCD_ILI9327_SB_DATA | func->pdata[i];
				++count;
			}
		}
		/* send command and data */
		ret = spi_write(&spi_config_w, &buf[0], count);
	}

	return ret;
}


/*******************/
/*@ LCDC operation */
/*******************/

/*-----------------------------------------------------------------------------
 * Function : lcd_tovis_ili9327_power_on
 * Abstruct : Power on the LCD
 * Input    : config : pointer to LCDC configure[OUT]
 * Return   : 0(success) / others(NG)
 *----------------------------------------------------------------------------*/
static int lcd_tovis_ili9327_power_on(const lcdc_config_t* config)
{
	int i;
	int ret;
	unsigned int reg;

	/* drive strength */
	reg = readl(CHG_DRIVE0);
	reg = (reg & ~0x00000003) | 0x00000001;
	writel(reg, CHG_DRIVE0);

	/* SPI pin init */
	reg = readl(CHG_PINSEL_G032);
	reg = (reg & ~0x78000000);
	writel(reg, CHG_PINSEL_G032);

	reg = readl(CHG_PULL7);
	reg = (reg & ~0x0FFFF000);
	writel(reg, CHG_PULL7);

	/* device reset */
	gpio_set_value(GPIO_LCD_XRST, 0x1);
	mdelay(1);
	gpio_set_value(GPIO_LCD_XRST, 0x0);
	mdelay(1);
	gpio_set_value(GPIO_LCD_XRST, 0x1);

	/* LCDC power on(supply clock) */
	lcdc_hw_init();

	/* LCDC initialize */
	lcdc_hw_set(config);

	/* start clock */
	lcdc_hw_bussel(LCDC_BUSSEL_BLACK);
	lcdc_hw_start();

	/* H/W power on */
	spi_probe(&spi_config_w);
	mdelay(10);

	for (i = 0; i < SPI_CMD_POWER_ON_NUM; i++) {
		ret = lcd_ili9327_send_command(&spi_cmd_power_on[i]);
		if (ret <= 0) {
			printf("ILI9327: spi_write error(%d)\n", ret);
			return ret;
		}
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 * Function : lcd_tovis_ili9327_power_off
 * Abstruct : Power off the LCD
 * Input    : none
 * Return   : none
 *----------------------------------------------------------------------------*/
static void lcd_tovis_ili9327_power_off(void)
{
	int i;
	unsigned int reg;

	/* write black data */
	lcdc_hw_bussel(LCDC_BUSSEL_BLACK);

	/* H/W power off */
	spi_probe(&spi_config_w);
	for (i = 0; i < SPI_CMD_POWER_OFF_NUM; i++) {
		lcd_ili9327_send_command(&spi_cmd_power_off[i]);
	}

	mdelay(40);
	spi_remove(&spi_config_w);

	/* SPI low */
	reg = readl(CHG_PINSEL_G032);
	reg = (reg | 0x78000000);
	writel(reg, CHG_PINSEL_G032);

	reg = readl(CHG_PULL7);
	reg = (reg & ~0x0FFFF000) | 0x03333000;
	writel(reg, CHG_PULL7);

	/* reset LCDC,IMC */
	lcdc_hw_stop();
	lcdc_hw_exit();

	/* reset the device */
	gpio_set_value(GPIO_LCD_XRST, 0x0);
}

#endif /* CONFIG_WM */
