/*
 * icx_lcd_tovis_ILI9341DS.c
 *
 * Copyright 2011-2014 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_WM

#include <asm/errno.h>

#include "reg.h"
#include "icx_io.h"
#include "icx_smu.h"
#include "icx_gpio.h"
#include "icx_spi.h"
#include "icx_pwc.h"
#include "icx_pwm.h"
#include "icx_lcdc.h"
#include "icx_lcd.h"


/*******************/
/*@ LCDC operation */
/*******************/

static int  lcd_tovis_ili9341ds_power_on(const lcdc_config_t* config);
static void lcd_tovis_ili9341ds_power_off(void);

const lcd_ops_t lcd_tovis_ili9341ds_ops = {
	.power_on  = lcd_tovis_ili9341ds_power_on,
	.power_off = lcd_tovis_ili9341ds_power_off,
};

/*****************/
/*@ LCDC setting */
/*****************/

/* hsync */
#define LCD_ILI9341DS_HAREA		LCDC_WIDTH_QVGA
#define	LCD_ILI9341DS_HPULSE		10
#define	LCD_ILI9341DS_HFRONTP		10
#define	LCD_ILI9341DS_HBACKP		20
#define	LCD_ILI9341DS_HBLANK		(LCD_ILI9341DS_HFRONTP + LCD_ILI9341DS_HPULSE + LCD_ILI9341DS_HBACKP)
#define	LCD_ILI9341DS_HTOTAL		(LCD_ILI9341DS_HAREA   + LCD_ILI9341DS_HBLANK)
#define	LCD_ILI9341DS_HEDGE1		(LCD_ILI9341DS_HFRONTP)
#define	LCD_ILI9341DS_HEDGE2		(LCD_ILI9341DS_HFRONTP + LCD_ILI9341DS_HPULSE)

/* vsync */
#define LCD_ILI9341DS_VAREA		LCDC_HEIGHT_QVGA
#define	LCD_ILI9341DS_VPULSE		2
#define	LCD_ILI9341DS_VFRONTP		4
#define	LCD_ILI9341DS_VBACKP		2
#define	LCD_ILI9341DS_VBLANK		(LCD_ILI9341DS_VFRONTP + LCD_ILI9341DS_VPULSE + LCD_ILI9341DS_VBACKP)
#define	LCD_ILI9341DS_VTOTAL		(LCD_ILI9341DS_VAREA   + LCD_ILI9341DS_VBLANK)
#define	LCD_ILI9341DS_VEDGE1		(LCD_ILI9341DS_VFRONTP)
#define	LCD_ILI9341DS_VEDGE2		(LCD_ILI9341DS_VFRONTP + LCD_ILI9341DS_VPULSE)

/* default brightness */
#define	LCD_ILI9341DS_PWM_DEFAULT	(PWM_TOTAL_100K / 2)

/* LCDC config */
const lcdc_config_t lcdc_tovis_ili9341ds_config = {
	.width	  = LCDC_WIDTH_QVGA,
	.height   = LCDC_HEIGHT_QVGA,

	.clkdiv   = SMU_PLLSEL_PLL3 | SMU_DIV(41),

	.frameadr = LCDC_FRAMEADR_DEFAULT,
	.areaadr  = LCDC_AREAADR_DEFAULT,
	.pol      = (LCDC_CLKPOL_NEG | LCDC_HPOL_POS | LCDC_VPOL_POS | LCDC_ENPOL_HI),
	.htotal   = LCD_ILI9341DS_HTOTAL,
	.harea	  = LCD_ILI9341DS_HAREA,
	.hedge1   = LCD_ILI9341DS_HEDGE1,
	.hedge2   = LCD_ILI9341DS_HEDGE2,
	.vtotal   = LCD_ILI9341DS_VTOTAL,
	.varea    = LCD_ILI9341DS_VAREA,
	.vedge1   = LCD_ILI9341DS_VEDGE1,
	.vedge2   = LCD_ILI9341DS_VEDGE2,

	.backlight = {
		.dev  = PWM_DEV_1,
		.pol  = LCDC_BLPOL_NEG,
		.full = LCD_ILI9341DS_PWM_DEFAULT,
		.div  = SMU_PLLSEL_PLL3 | SMU_DIV(4),
	},
};

/****************/
/*@ SPI setting */
/****************/

/* SPI config */
static spi_config_t spi_config_w = {
	.dev = SPI_DEV_SP3,
	.clk = SPI_CLK_6MHZ,
	.cs  = SPI_CS_SEL_CS0,
	.pol = (SPI_CK_DLY_ON | SPI_CK_POL_POS | SPI_CS_POL_NEG),
	.nbw = SPI_NB_9BIT,
	.nbr = 0,
};

/* SPI start byte */
#define LCD_ILI9341DS_SB_CMD	0x0000          /* Set index register                           */
#define LCD_ILI9341DS_SB_DATA    0x0100          /* Writes instruction or RAM data       */

#define LCD_ILI9341DS_IR_SYNC	0xFFFE		/* transfer synchronize(local rule)	*/
#define LCD_ILI9341DS_IR_WAIT	0xFFFF		/* wait command(local rule)			*/

#define	LCD_ILI9341DS_HEART_BEAT_MAGIC_NUMBER	0xAA55

/* command type */
typedef struct lcd_ili9341ds_command {
	const unsigned short    ir;
	union {
		const unsigned short    data_size;
		const unsigned short    wait_time_ms;
	};
	const char* const       pdata;
} lcd_ili9341ds_command_t;

/* SPI command */
static const lcd_ili9341ds_command_t spi_cmd_power_on[] = {
	/* Power Supply Setting */
	{ 0xC0, { 1 }, "\x23" },
	{ 0xC1, { 1 }, "\x10" },
	{ 0xC5, { 2 }, "\x30\x4A" },
	{ 0xCB, { 5 }, "\x39\x2C\x00\x34\x02" },
	{ 0xCF, { 3 }, "\x00\x89\x30" },

	/* Driver timing control */
	{ 0xEA, { 2 }, "\x00\x00" },

	/* Power on sequence control */
	{ 0xED, { 4 }, "\x64\x03\x12\x81" },

	/* Interface control */
	{ 0xF6, { 3 }, "\x01\x00\x06" },

	/* Pump ratio control */
	{ 0xF7, { 1 }, "\x20" },

	/* Display mode setting */
	{ 0x36, { 1 }, "\x48" },
	{ 0x3A, { 1 }, "\x66" },
	{ 0xB0, { 1 }, "\xC2" },
	{ 0xB4, { 1 }, "\x00" },
	{ 0xB5, { 4 }, "\x04\x02\x0A\x14" },
	{ 0xB6, { 4 }, "\x0A\x02\x27\x04" },

	/* gamma settings */
	{ 0xE0, { 15 }, "\x06\x2C\x24\x0A\x0B\x03\x44\x55\x32\x04\x10\x04\x11\x11\x00" }, /* p1 - p15 */
	{ 0xE1, { 15 }, "\x0A\x0D\x16\x05\x10\x04\x36\x24\x4C\x07\x0F\x0D\x2D\x2E\x17" }, /* p1 - p15 */

	/* exit sleep */
	{ 0x11, { 0 }, NULL },
	{ LCD_ILI9341DS_IR_WAIT, { 80 }, NULL },

	/* Display On */
	{ LCD_ILI9341DS_IR_WAIT, { 80 }, NULL },
	{ 0x29, { 0 }, NULL },
};
#define SPI_CMD_POWER_ON_NUM    ARRAY_SIZE(spi_cmd_power_on)

#define SPI_CMD_DATA_MAX_SIZE   (1 + 15) /* command(1) + data(15) */

static const lcd_ili9341ds_command_t spi_cmd_power_off[] = {

	/* Display Off */
	{ 0x28, { 0 }, NULL },
	{ LCD_ILI9341DS_IR_WAIT, { 50 }, NULL },

	 /* sleep in */
	{ 0x10, { 0 }, NULL },
	{ LCD_ILI9341DS_IR_WAIT, { 120 }, NULL },
};

#define SPI_CMD_POWER_OFF_NUM   ARRAY_SIZE(spi_cmd_power_off)

/******************/
/*@ Local routine */
/******************/

/*-----------------------------------------------------------------------------
 * Function : lcd_ili9341ds_send_command
 * Abstruct : Send a command & a data to LCD via SPI
 * Input    : none
 * Return   : <= 0(NG) / others(sent packet=success)
 *----------------------------------------------------------------------------*/
static int lcd_ili9341ds_send_command(const lcd_ili9341ds_command_t* func)
{
	int count = 0;
	int ret   = 0;
	u_int buf[SPI_CMD_DATA_MAX_SIZE] = { 0x0 };

	if (func->ir == (unsigned short)LCD_ILI9341DS_IR_WAIT) {
		/* wait */
		mdelay(func->wait_time_ms);
		ret = func->wait_time_ms;
	}
	else {
		/* make command and data */
		buf[count] = LCD_ILI9341DS_SB_CMD | func->ir;
		++count;
		if (func->data_size && func->pdata) {

			int i;

			if (SPI_CMD_DATA_MAX_SIZE < func->data_size) {
				printf("%s(): error -1\n", __FUNCTION__);

				return -1;
			}

			for (i = 0; i < func->data_size; ++i) {
				buf[count] = LCD_ILI9341DS_SB_DATA | func->pdata[i];
				++count;
			}
		}
		/* send command and data */
		ret = spi_write(&spi_config_w, &buf[0], count);
	}

	return ret;
}


/*******************/
/*@ LCDC operation */
/*******************/

/*-----------------------------------------------------------------------------
 * Function : lcd_tovis_ili9341ds_power_on
 * Abstruct : Power on the LCD
 * Input    : config : pointer to LCDC configure[OUT]
 * Return   : 0(success) / others(NG)
 *----------------------------------------------------------------------------*/
static int lcd_tovis_ili9341ds_power_on(const lcdc_config_t* config)
{
	int i;
	int ret;

	/* device reset */
	gpio_set_value(GPIO_LCD_XRST, 0x1);
	mdelay(1);
	gpio_set_value(GPIO_LCD_XRST, 0x0);
	mdelay(1);
	gpio_set_value(GPIO_LCD_XRST, 0x1);

	mdelay(120);

	/* LCDC power on(supply clock) */
	lcdc_hw_init();

	/* LCDC initialize */
	lcdc_hw_set(config);

	/* start clock */
	lcdc_hw_bussel(LCDC_BUSSEL_BLACK);
	lcdc_hw_start();
	mdelay(10);

	/* H/W power on */
	spi_probe(&spi_config_w);

	for (i = 0; i < SPI_CMD_POWER_ON_NUM; i++) {
		ret = lcd_ili9341ds_send_command(&spi_cmd_power_on[i]);
		if (ret <= 0) {
			printf("ILI9341DS: spi_write error(%d)\n", ret);
			return ret;
		}
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 * Function : lcd_tovis_ili9341ds_power_off
 * Abstruct : Power off the LCD
 * Input    : none
 * Return   : none
 *----------------------------------------------------------------------------*/
static void lcd_tovis_ili9341ds_power_off(void)
{
	int i;

	/* write black data */
	lcdc_hw_bussel(LCDC_BUSSEL_BLACK);

	mdelay(50);

	/* H/W power off */
	spi_probe(&spi_config_w);
	for (i = 0; i < SPI_CMD_POWER_OFF_NUM; i++) {
		lcd_ili9341ds_send_command(&spi_cmd_power_off[i]);
	}

	spi_remove(&spi_config_w);

	/* reset LCDC,IMC */
	lcdc_hw_stop();
	lcdc_hw_exit();

	/* reset the device */
	gpio_set_value(GPIO_LCD_XRST, 0x0);

	mdelay(120);
}

#endif /* CONFIG_WM */
