/*
 * icx_lcd_tovis_R61525.c
 *
 * Copyright 2011-2015 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_WM

#include <asm/errno.h>

#include "reg.h"
#include "icx_io.h"
#include "icx_smu.h"
#include "icx_gpio.h"
#include "icx_spi.h"
#include "icx_pwc.h"
#include "icx_pwm.h"
#include "icx_lcdc.h"
#include "icx_lcd.h"

/*******************/
/*@ LCDC operation */
/*******************/

static int  lcd_tovis_r61525_power_on(const lcdc_config_t* config);
static void lcd_tovis_r61525_power_off(void);

const lcd_ops_t lcd_tovis_r61525_ops = {
	.power_on		   = lcd_tovis_r61525_power_on,
	.power_off		   = lcd_tovis_r61525_power_off,
};

/*****************/
/*@ LCDC setting */
/*****************/

/* hsync */
#define LCD_R61525_HAREA		LCDC_WIDTH_WQVGA
#define	LCD_R61525_HPULSE		12
#define	LCD_R61525_HFRONTP		12
#define	LCD_R61525_HBACKP		20
#define	LCD_R61525_HBLANK		(LCD_R61525_HFRONTP + LCD_R61525_HPULSE + LCD_R61525_HBACKP)
#define	LCD_R61525_HTOTAL		(LCD_R61525_HAREA + LCD_R61525_HBLANK)
#define	LCD_R61525_HEDGE1		(LCD_R61525_HFRONTP)
#define	LCD_R61525_HEDGE2		(LCD_R61525_HFRONTP + LCD_R61525_HPULSE)

/* vsync */
#define LCD_R61525_VAREA		LCDC_HEIGHT_WQVGA
#define	LCD_R61525_VPULSE		4
#define	LCD_R61525_VFRONTP		8
#define	LCD_R61525_VBACKP		8
#define	LCD_R61525_VBLANK		(LCD_R61525_VFRONTP + LCD_R61525_VPULSE + LCD_R61525_VBACKP)
#define	LCD_R61525_VTOTAL		(LCD_R61525_VAREA + LCD_R61525_VBLANK)
#define	LCD_R61525_VEDGE1		(LCD_R61525_VFRONTP)
#define	LCD_R61525_VEDGE2		(LCD_R61525_VFRONTP + LCD_R61525_VPULSE)

/* default brightness */
#define	LCD_R61525_PWM_DEFAULT	(PWM_TOTAL_100K / 2)

/* LCDC config */
lcdc_config_t lcdc_tovis_r61525_config = {
	.width	  = LCD_R61525_HAREA,
	.height   = LCD_R61525_VAREA,

	.clkdiv   = SMU_PLLSEL_PLL3 | SMU_DIV(32),

	.frameadr = LCDC_FRAMEADR_DEFAULT,
	.areaadr  = LCDC_AREAADR_DEFAULT,
	.pol	  = (LCDC_CLKPOL_POS | LCDC_HPOL_POS | LCDC_VPOL_POS | LCDC_ENPOL_HI),	// C6h register(Interface Setting)

	.htotal   = LCD_R61525_HTOTAL, 
	.harea	  = LCD_R61525_HAREA,
	.hedge1   = LCD_R61525_HEDGE1, 
	.hedge2   = LCD_R61525_HEDGE2,
	.vtotal   = LCD_R61525_VTOTAL,
	.varea    = LCD_R61525_VAREA,
	.vedge1   = LCD_R61525_VEDGE1,
	.vedge2   = LCD_R61525_VEDGE2,

	.backlight = {
		.dev  = PWM_DEV_1,
		.pol  = LCDC_BLPOL_NEG,
		.full = LCD_R61525_PWM_DEFAULT,
		.div  = SMU_PLLSEL_PLL3 | SMU_DIV(4),
	},
};

/****************/
/*@ SPI setting */
/****************/

/* SPI config */
static spi_config_t spi_config_w = { 
	.dev = SPI_DEV_SP3,
	.clk = SPI_CLK_1MHZ,
	.cs  = SPI_CS_SEL_CS0,
	.pol = (SPI_CK_DLY_OFF | SPI_CK_POL_NEG | SPI_CS_POL_NEG),
	.nbw = SPI_NB_9BIT,	/* 9bit per send */
	.nbr = 0,
};

/* SPI start byte */
#define LCD_R61525_SB_CMD	0x0000		/* Set index register				*/
#define LCD_R61525_SB_DATA	0x0100		/* Writes instruction or RAM data	*/

#define LCD_R61525_IR_SYNC	0xFFFE		/* transfer synchronize(local rule)	*/
#define LCD_R61525_IR_WAIT	0xFFFF		/* wait command(local rule)			*/

/* command type */
typedef struct lcd_r61525_command {
	const unsigned short	ir;
	union {
		const unsigned short	data_size;
		const unsigned short	wait_time_ms;
	};
	const char* const	pdata;
} lcd_r61525_command_t;

/* SPI command */
static const lcd_r61525_command_t spi_cmd_power_on[] = {
	{ 0xB0, { 1 }, "\x00" },				/* Manufacturer Command Access Protect Off */
	{ 0xB3, { 4 }, "\x00\x00\x00\x00" },	/* Frame memory Access and Interface setting */
	{ 0xB4, { 1 }, "\x11" },				/* Display mode and Frame memory Write mode setting */
											/* DPI */
	{ 0xC0, { 8 }, "\x03\x63\x00\x00\x00\x00\x01\x43" },	/* Panel Driving setting */
															/* BLV? */

	{ 0xC1, { 4 }, "\x08\x14\x08\x08" },	/* Display Timing Setting for Normal Mode */
	{ 0xC2, { 4 }, "\x08\x14\x08\x08" },	/* Display Timing Setting for Partial Mode */
	{ 0xC3, { 4 }, "\x08\x14\x08\x08" },	/* Display Timing Setting for Idle Mode */
											/* BP,FP = 8lines */
	{ 0xC4, { 4 }, "\x11\x01\x13\x01" }, /* Source/VCOM/Gate Driving Timing Setting */

	{ 0xC6, { 1 }, "\x03" },	/* Interface Setting */
								/* VSYNC:4 = Low active, HSYNC:3 = Low Active */
								/* ENABLE:1 = High active, DOTCLK:0 = falling */

	/* Gamma Set A-C */
	{ 0xC8, { 22 }, "\x04\x04\x15\x13\x16\x0E\x0B\x11\x02\x05\x00\x06\x05\x0C\x0E\x18\x12\x15\x05\x04\x0B\x00" },
	{ 0xC9, { 22 }, "\x04\x04\x15\x13\x16\x0E\x0B\x11\x02\x05\x00\x06\x05\x0C\x0E\x18\x12\x15\x05\x04\x0B\x00" },
	{ 0xCA, { 22 }, "\x04\x04\x15\x13\x16\x0E\x0B\x11\x02\x05\x00\x06\x05\x0C\x0E\x18\x12\x15\x05\x04\x0B\x00" },

	{ 0xD2, { 3 }, "\x03\x24\x04" },	/* Power Setting for Normal Mode */
	{ 0xD3, { 3 }, "\x03\x24\x04" },	/* Power Setting for Partial Mode */
	{ 0xD4, { 3 }, "\x03\x24\x04" },	/* Power Setting for Idle Mode */

	{ 0xB0, { 1 }, "\x00" },	/* Manufacturer Command Access Protect on */

	{ 0x3A, { 1 }, "\x60" },	/* Set pixel format */
								/* DPI = 18bits */

	{ 0x11, { 0 }, NULL },	/* exit sleep */

	{ LCD_R61525_IR_WAIT, { 120 }, NULL },

	{ 0x34, { 0 }, NULL },	/* set tear off */
	{ 0x36, { 1 }, "\x00" },	/* set address mode */

	{ 0x2A, { 4 }, "\x00\x00\x00\xEF" },	/* Set column address : 240 */
	{ 0x2B, { 4 }, "\x00\x00\x01\x8F" },	/* Set page address : 400 */

	{ 0xD0, { 4 }, "\x07\x07\x2F\x66" },	/* Power Setting */
	{ 0xD1, { 3 }, "\x03\x69\x6D" },		/* VCOM Setting */

	{ 0x29, { 0 }, NULL },	/* Display On */
	{ 0x2C, { 0 }, NULL },	/* write_memory_start */
};
#define	SPI_CMD_POWER_ON_NUM	ARRAY_SIZE(spi_cmd_power_on)

#define	SPI_CMD_DATA_MAX_SIZE	(1 + 22)	/* command(1) + data(22) */

static const lcd_r61525_command_t spi_cmd_power_off[] = {
	{ 0x28, { 0 }, NULL },	/* Display Off */
	{ LCD_R61525_IR_WAIT, { 40 }, NULL },
	{ 0x10, { 0 }, NULL },	/* sleep in */
	{ LCD_R61525_IR_WAIT, { 120 }, NULL },
	{ 0xB4, { 1 }, "\x00" },
	{ 0xB1, { 1 }, "\x01" },
};
#define	SPI_CMD_POWER_OFF_NUM	ARRAY_SIZE(spi_cmd_power_off)

/******************/
/*@ local routine */
/******************/

/*-----------------------------------------------------------------------------
 *  Abstruct    : Send a command & a data to LCD via SPI
 *  Input       : none
 *  Return      : <= 0(NG) / others(sent packet=success)
 *----------------------------------------------------------------------------*/
static int lcd_r61525_send_command(const lcd_r61525_command_t* func)
{
	int count = 0;
	int ret = 0;

	u_int buf[SPI_CMD_DATA_MAX_SIZE] = { 0x0 };

	if (func->ir == (unsigned short)LCD_R61525_IR_WAIT) {
		/* wait */
		mdelay(func->wait_time_ms);
		ret = func->wait_time_ms;
	}
	else {
		/* make command and data */
		buf[count] = LCD_R61525_SB_CMD | func->ir;
		++count;
		if (func->data_size && func->pdata) {

			int i;

			if (SPI_CMD_DATA_MAX_SIZE < func->data_size) {
				printf("%s(): error -1\n", __FUNCTION__);

				return -1;
			}

			for (i = 0; i < func->data_size; ++i) {
				buf[count] = LCD_R61525_SB_DATA | func->pdata[i];
				++count;
			}
		}
		/* send command and data */
		ret = spi_write(&spi_config_w, &buf[0], count);
//		ret = spi_write_tiecs(&spi_config_w, &buf[0], count);
	}

	return ret;
}


/*******************/
/*@ LCDC operation */
/*******************/
#define	CHG_DRIVE1_LCD_MSK	(0x00000030)
#define	CHG_DRIVE1_LCD_4MA_VAL	(0x00000010)

static void change_rgbif_drive_capability_to_4mA(void)
{
	u_int regval = inl(CHG_DRIVE1) & ~CHG_DRIVE1_LCD_MSK;
	outl((regval | CHG_DRIVE1_LCD_4MA_VAL), CHG_DRIVE1);
}

/*-----------------------------------------------------------------------------
 *  Abstruct    : Power on the LCD 
 *  Input       : config : pointer to LCDC configure[OUT]
 *  Return      : 0(success) / others(NG)
 *----------------------------------------------------------------------------*/
static int lcd_tovis_r61525_power_on(const lcdc_config_t* config)
{
	int i;
	int ret;

	change_rgbif_drive_capability_to_4mA();

	/* device reset */
	gpio_set_value(GPIO_LCD_XRST, 0x0);
	udelay(1000);
	gpio_set_value(GPIO_LCD_XRST, 0x1);

	udelay(5000);

	/* LCDC power on(supply clock) */
	lcdc_hw_init();

	/* LCDC initialize */
	lcdc_hw_set(config);

	/* H/W power on */
	spi_probe(&spi_config_w);
	for (i = 0; i < SPI_CMD_POWER_ON_NUM; i++) {
		ret = lcd_r61525_send_command(&spi_cmd_power_on[i]);
		if (ret <= 0) {
			printf("R61525: spi_write error(%d)\n", ret);
			return ret;
		}
	}

	/* start clock */
	lcdc_hw_bussel(LCDC_BUSSEL_BLACK);
	lcdc_hw_start();
	udelay(10000);

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Abstruct    : Power off the LCD 
 *  Input       : none
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void lcd_tovis_r61525_power_off(void)
{
	int i;

	/* write black data */
	lcdc_hw_bussel(LCDC_BUSSEL_BLACK);

	mdelay(120);

	/* H/W power off */
	spi_probe(&spi_config_w);
	for (i = 0; i < SPI_CMD_POWER_OFF_NUM; i++) {
		lcd_r61525_send_command(&spi_cmd_power_off[i]);
	}

	/* reset the device */
	gpio_set_value(GPIO_LCD_XRST, 0x0);

	/* reset LCDC,IMC */
	lcdc_hw_stop();
	lcdc_hw_exit();
}


#endif	/* CONFIG_WM */ 
